#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <immintrin.h>
#include <sys/time.h>

#include "TablaHash.h"

int main(int argc, char* argv[]) {
	int i;
	int numH;
	struct timeval horaI, horaF;
	unsigned int usegs1, usegs2;
	unsigned int timeResults[2];

	if (argc!=6 && argc!=7) {
		printf("5 parameters needed: NumHilos NumIter NumEntr TamTabla outputFile [verbose]\n");
		exit(-1);
	}

	numH = atoi(argv[1]);
	int MAXITER = atoi(argv[2]);
	int NUMENTRADAS = atoi(argv[3]);
	int TAMTABLAHASH = atoi(argv[4]);
	FILE *f = fopen(argv[5], "a");
	if (argc > 6) {
		verbose = 1;
	}

	if (f == NULL) {
		printf("No se pudo abrir el fichero.\n");
		exit(1);
	}

	if (verbose) {
		printf("Calentamiento...\n"); fflush(stdout);
	}
	init_tabla(NUMENTRADAS, TAMTABLAHASH);
	if (verbose) {
		printf("done.\n"); fflush(stdout);
	}

	// TRANSACCIONAL = 1; => activada
	// TRANSACCIONAL = 0; => no activada
	//Bucle principal
	int TRANSACCIONAL;
	for (TRANSACCIONAL=0; TRANSACCIONAL<2; TRANSACCIONAL++) {
		reset_thread_counter();

		global_params.tam_tabla_hash = TAMTABLAHASH;
		global_params.num_entradas = NUMENTRADAS;
		global_params.max_iteraciones = FIXED_WORKLOAD ? MAXITER/(2*numH) : MAXITER;
		global_params.use_tsx = TRANSACCIONAL;

		// hora inicial
		gettimeofday(&horaI, NULL);

		for (i=0; i<numH; i++) {
			crea_hilo_busca();
		}

		for (i=numH; i<2*numH; i++) {
			crea_hilo_modifica();
		}

		join_all_threads();

		// hora final
		gettimeofday(&horaF, NULL);

		horaF.tv_sec = horaF.tv_sec - horaI.tv_sec;
		horaI.tv_sec = 0;
		usegs1 = horaI.tv_usec;
		usegs2 = horaF.tv_sec * 1000000 + horaF.tv_usec;

		timeResults[TRANSACCIONAL] = usegs2 - usegs1;
	}

	destroy_attr();

	struct tsxStats stats = getTsxStats();

	fprintf(f, "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
			numH,
			MAXITER,
			NUMENTRADAS,
			TAMTABLAHASH,
			FIXED_WORKLOAD ? MAXITER : MAXITER*numH*2,
			stats.ok,
			stats.nok,
			stats.capacity,
			stats.conflict,
			stats.explicit_tsx,
			stats.retriable,
			stats.debug,
			stats.nested,
			timeResults[1],
			timeResults[0]
	);

	if (verbose) {
		printf("%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
			   numH,
			   MAXITER,
			   NUMENTRADAS,
			   TAMTABLAHASH,
			   FIXED_WORKLOAD ? MAXITER : MAXITER*numH*2,
			   stats.ok,
			   stats.nok,
			   stats.capacity,
			   stats.conflict,
			   stats.explicit_tsx,
			   stats.retriable,
			   stats.debug,
			   stats.nested,
			   timeResults[1],
			   timeResults[0]
		);
	}
	return 0;
}


void init_tabla(int num_entradas, int tam_tabla_hash) {
	int scope, detach_state;
	int i;

	pthread_mutex_init(&mutex, NULL);
	pthread_attr_init(&attr);

	if (verbose) {
		pthread_attr_getscope(&attr, &scope);
		pthread_attr_getdetachstate(&attr, &detach_state);
		printf("Scope               = %s\n",
			   (scope == PTHREAD_SCOPE_SYSTEM)  ? "PTHREAD_SCOPE_SYSTEM"  :
			   (scope == PTHREAD_SCOPE_PROCESS) ? "PTHREAD_SCOPE_PROCESS" :
				"???");

		printf("Detach state        = %s\n",
			   (detach_state == PTHREAD_CREATE_DETACHED) ? "PTHREAD_CREATE_DETACHED" :
			   (detach_state == PTHREAD_CREATE_JOINABLE) ? "PTHREAD_CREATE_JOINABLE" :
			   "???");
	}

	Hash = (EXT**) calloc(num_entradas, sizeof(EXT *));
	for (i = 0; i < num_entradas; i++) {
		Hash[i] = NULL;
	}

	float entrada;
	while (!caliente) {
		entrada = (float) rand() * (float) num_entradas / (float) RAND_MAX;
		rellenar_tabla((int) entrada, tam_tabla_hash, num_entradas);
	}
}

void rellenar_tabla(int entrada, int tam_tabla_hash, int num_entradas) {
	EXT *pp;
	int last = 0;

	pp = Hash[entrada];
	if (pp!=NULL) {
		while (pp->next!=NULL) {
			pp = pp->next;
		}
		last = pp->valor;
		if (last+1 == tam_tabla_hash) {
			nentradas++;
		}
		if (nentradas >= num_entradas) {
			caliente=1;
		}
		if (last < tam_tabla_hash) { // Limite extensiones por entrada
			pp->next = (EXT *) calloc(1, sizeof(EXT));
			pp = pp->next;
			pp->valor = last+1;
			pp->modifica = 0;
			pp->next=NULL;
		}
	} else { // Primer elemento
		Hash[entrada] = (EXT *) calloc(1, sizeof(EXT));
		pp = Hash[entrada];
		pp->valor = last;
		pp->modifica = 0;
		pp->next=NULL;
	}
}

void acquire_lock(pthread_mutex_t *mutex, int use_tsx) {
	unsigned int status;

	if (use_tsx) {
		if ((status = _xbegin()) == _XBEGIN_STARTED) {
			if ((*mutex).__data.__lock == 1) {
				_xend();
				pthread_mutex_lock(mutex);
			}
		} else {
			if (status & _XABORT_RETRY)
				retry++;
			if (status & _XABORT_CONFLICT)
				conflict++;
			if (status & _XABORT_CAPACITY)
				capacity++;
			if (status & _XABORT_DEBUG)
				debug++;
			if (status & _XABORT_NESTED)
				nested++;
			if (status & _XABORT_EXPLICIT)
				explicitTran++;

			pthread_mutex_lock(mutex);
		}
	} else {
		pthread_mutex_lock(mutex);
	}
}

void release_lock(pthread_mutex_t *mutex, int use_tsx) {
	if (use_tsx) {
		if ((*mutex).__data.__lock == 0) {
			_xend();
			nTranOK++;
		} else {
			nTranNOK++;
			pthread_mutex_unlock(mutex);
		}
	} else {
		pthread_mutex_unlock(mutex);
	}
}

void reset_thread_counter() {
	thread_count = 0;
}

static void * hiloModifica() {
	int i, entrada, num;

	for (i=0; i < global_params.max_iteraciones; i++) {
		entrada = (float) rand() * (float) global_params.num_entradas / (float) RAND_MAX;
		num = (float) rand()* (float) global_params.tam_tabla_hash / (float) RAND_MAX;
		Modifica((int) entrada, (int) num, global_params.use_tsx);
	}

	pthread_exit(NULL);
}

void crea_hilo_modifica() {
	pthread_create(&tid[thread_count++], &attr, &hiloModifica, NULL);
}

void Modifica(int Entrada, int num, int use_tsx) {
	EXT *pp;
	int fin = 0;

	acquire_lock(&mutex, use_tsx);
	pp = Hash[Entrada];
	while ( (pp!=NULL) && (fin==0) ) {
		if (pp->valor == num) {
			fin = 1;
			pp->modifica++; // Escritura para provocar conflictos de datos.
		}
		pp = pp->next;
	}
	release_lock(&mutex, use_tsx);
}

static void * hiloBusca() {
	int i, entrada, num;

	for (i=0; i < global_params.max_iteraciones; i++) {
		entrada = (float) rand() * (float) global_params.num_entradas / (float) RAND_MAX;
		num = (float) rand() * (float) global_params.tam_tabla_hash / (float) RAND_MAX;
		Busca((int) entrada, (int) num, global_params.use_tsx);
	}

	pthread_exit(NULL);
}

void crea_hilo_busca() {
	pthread_create(&tid[thread_count++], &attr, &hiloBusca, NULL);
}

void Busca(int Entrada, int num, int use_tsx) {
	EXT *pp;
	int fin = 0;

	acquire_lock(&mutex, use_tsx);
	pp = Hash[Entrada];
	while ( (pp!=NULL) && (fin==0) ) {
		if (pp->valor == num)
			fin = 1;
		pp = pp->next;
	}
	release_lock(&mutex, use_tsx);
}

void join_all_threads() {
	int i;
	for (i = 0; i < thread_count; i++) {
		pthread_join(tid[i], &res);
	}
}

struct tsxStats getTsxStats() {
	struct tsxStats stats;
	stats.retriable = retry;
	stats.capacity = capacity;
	stats.conflict = conflict;
	stats.debug = debug;
	stats.explicit_tsx = explicitTran;
	stats.nested = nested;
	stats.nok = nTranNOK;
	stats.ok = nTranOK;

	return stats;
}

void destroy_attr() {
	pthread_attr_destroy(&attr);
}