#include <pthread.h>
#include <stdio.h>
#include <immintrin.h>
#include <stdlib.h>
#include <sys/time.h>

#define MAXT 256

typedef struct extension {
	int valor;
	int modifica;
	struct extension *next;
} EXT;
EXT **Hash;

pthread_mutex_t mutex;
int nTranOK = 0;
int nTranNOK = 0;
int retry = 0;
int conflict = 0;
int capacity = 0;
int debug = 0;
int nested = 0;
int explicitTran = 0;

int nentradas = 0, caliente = 0;

// Parametrización
int MAXITER;
int TAMTABLAHASH;
int NUMENTRADAS;
int TRANSACCIONAL;

void InsertaCalienta(int Entrada);
void Calienta();
void Modifica(int Entrada, int num);
void acquire_lock(pthread_mutex_t *mutex);
void release_lock(pthread_mutex_t *mutex);
void Busca(int Entrada, int num);
static void * hiloBusca();
static void * hiloModifica();

int main(int argc, char* argv[]) {
	int i, s;
	int numH;
	int verbose;
	pthread_t tid[MAXT];
	pthread_attr_t attr;
	void *res;
	struct timeval horaI, horaF;
	unsigned int usegs1, usegs2;
	unsigned int timeResults[2];

	if (argc!=6) {
		printf("5 parameters needed: NumHilos NumIter NumEntr TamTabla outputFile [verbose]\n"),
				exit(-1);
	}
	numH = atoi(argv[1]);
	MAXITER = atoi(argv[2]);
	NUMENTRADAS = atoi(argv[3]);
	TAMTABLAHASH = atoi(argv[4]);
	FILE *f = fopen(argv[5], "a");
	if (argc > 6) {
		verbose = 1;
	}

	if (f == NULL) {
		printf("No se pudo abrir el fichero.\n");
		exit(1);
	}

	//
	// Inis
	Hash = (EXT**) calloc(NUMENTRADAS, sizeof(EXT *));
	for (i=0; i<NUMENTRADAS; i++)
		Hash[i]=NULL;

	pthread_mutex_init(&mutex, NULL);
	s = pthread_attr_init(&attr);

	s = pthread_attr_getscope(&attr, &i);

	if (verbose) {
		if (i == PTHREAD_SCOPE_SYSTEM) {
			printf(" Hilos sistema.\n");
		} else if (i == PTHREAD_SCOPE_PROCESS) {
			printf(" Hilos proceso.\n");
		} else {
			printf(" Hilos desonocidos.");
		}
	}

	if (verbose) {
		printf("Calentamiento..."); fflush(stdout);
	}
	Calienta();
	if (verbose) {
		printf("done.\n"); fflush(stdout);
	}

	// TRANSACCIONAL = 1; => activada
	// TRANSACCIONAL = 0; => no activada
	// Bucle principal
	for (TRANSACCIONAL=0; TRANSACCIONAL<2; TRANSACCIONAL++) {
		if (TRANSACCIONAL==1) {
			if (verbose) {
				printf("Transactional. ");
				//  if ((status = _xtest()) != 0)
				//    printf("active.");
				//  else
				//    printf("not active.");
			}
		}

		// hora inicial
		gettimeofday(&horaI, NULL);

		for (i=0; i<numH; i++)
			pthread_create(&tid[i], &attr, &hiloBusca, NULL);
		for (i=numH; i<2*numH; i++)
			pthread_create(&tid[i], &attr, &hiloModifica, NULL);
		// pthread_create(&tid[i], &attr, &hiloBusca, NULL);

		s = pthread_attr_destroy(&attr);

		for (i=0; i<2*numH; i++)
			pthread_join(tid[i], &res);

		// hora final
		gettimeofday(&horaF, NULL);

		horaF.tv_sec = horaF.tv_sec - horaI.tv_sec;
		horaI.tv_sec = 0;
		usegs1 = horaI.tv_usec;
		usegs2 = horaF.tv_sec * 1000000 + horaF.tv_usec;

		timeResults[TRANSACCIONAL] = usegs2 - usegs1;
	}

	fprintf(f, "%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d\n", numH, MAXITER, NUMENTRADAS, TAMTABLAHASH, MAXITER*numH*2, nTranOK, nTranNOK, capacity, conflict, explicitTran, retry, debug, nested, timeResults[1], timeResults[0]);
	return 0;
}

/*

   if ((status = _xbegin()) == _XBEGIN_STARTED) {

    flag=1;
    _xend();

   } else {

    printf("tr nok\n");

    if (status == _XABORT_RETRY)
      printf("Retry possible\n");
    if (status == _XABORT_CONFLICT)
      printf("CONFLICT\n");
    if (status == _XABORT_CAPACITY)
      printf("CAPACITY\n");
    if (status == _XABORT_DEBUG)
      printf("DEBUG\n");
    if (status == _XABORT_NESTED)
      printf("NESTED\n");
    if (status == _XABORT_EXPLICIT)
      printf("EXPLICIT\n");

   printf("Status: 0x%x\n", status);

   }

   printf ("Flag: %d\n", flag);

   }
 */


void Calienta() {
	int entrada;
	float fentrada;

	while (!caliente) {
		fentrada = (float) rand() * (float) NUMENTRADAS / (float) RAND_MAX;
		entrada = (int) fentrada;
		InsertaCalienta(entrada);
	}
}

void InsertaCalienta(int Entrada) {
	EXT *pp;
	int last = 0;

	pp = Hash[Entrada];
	if (pp!=NULL) {
		while (pp->next!=NULL) {
			pp = pp->next;
		}
		last = pp->valor;
		if (last+1==TAMTABLAHASH) {
			nentradas++;
		}
		if (nentradas >= NUMENTRADAS) {
			caliente=1;
		}
		if (last < TAMTABLAHASH) { // Limite extensiones por entrada
			pp->next = (EXT *) calloc(1, sizeof(EXT));
			pp = pp->next;
			pp->valor = last+1;
			pp->modifica = 0;
			pp->next=NULL;
		}
	} else { // Primer elemento
		Hash[Entrada] = (EXT *) calloc(1, sizeof(EXT));
		pp = Hash[Entrada];
		pp->valor = last;
		pp->modifica = 0;
		pp->next=NULL;
	}
}

static void * hiloModifica() {
	int i, entrada, num;
	float fentrada, fnum;

	for (i=0; i< MAXITER; i++) {
		fentrada = (float) rand() * (float) NUMENTRADAS / (float) RAND_MAX;
		entrada = (int) fentrada;
		fnum = (float) rand()* (float) TAMTABLAHASH / (float) RAND_MAX;
		num = (int) fnum;
		Modifica(entrada, num);
	}
}

void Modifica(int Entrada, int num) {
	EXT *pp;
	int fin = 0;

	acquire_lock(&mutex);
	pp = Hash[Entrada];
	while ( (pp!=NULL) && (fin==0) ) {
		if (pp->valor == num) {
			fin = 1;
			pp->modifica++; // Write to create data conflicts
		}
		pp = pp->next;
	}
	release_lock(&mutex);
}

//
// Implementación transaccional de semáforos binarios
//
void acquire_lock(pthread_mutex_t *mutex) {
	unsigned int status;

	if (TRANSACCIONAL==1) {
		if ((status = _xbegin()) == _XBEGIN_STARTED) {
			if ((*mutex).__data.__lock == 1) {
				_xend();
				pthread_mutex_lock(mutex);
			}
		} else {
			nTranNOK++;
			if (status & _XABORT_RETRY)
				retry++;
			if (status & _XABORT_CONFLICT)
				conflict++;
			if (status & _XABORT_CAPACITY)
				capacity++;
			if (status & _XABORT_DEBUG)
				debug++;
			if (status & _XABORT_NESTED)
				nested++;
			if (status & _XABORT_EXPLICIT)
				explicitTran++;

			pthread_mutex_lock(mutex);
		}
	} else {
		pthread_mutex_lock(mutex);
	}
}

void release_lock(pthread_mutex_t *mutex) {
	if (TRANSACCIONAL==1) {
		if ((*mutex).__data.__lock == 0) {
			_xend();
			nTranOK++;
		} else {
			pthread_mutex_unlock(mutex);
		}
	} else {
		pthread_mutex_unlock(mutex);
	}
}

static void * hiloBusca() {
	int i, entrada, num;
	float fentrada, fnum;

	for (i=0; i< MAXITER; i++) {
		fentrada = (float) rand()* (float) NUMENTRADAS / (float) RAND_MAX;
		fnum = (float) rand()* (float) TAMTABLAHASH / (float) RAND_MAX;
		entrada = (int) fentrada;
		num = (int) fnum;
		Busca(entrada, num);
	}
}

void Busca(int Entrada, int num) {
	EXT *pp;
	int fin = 0;

	acquire_lock(&mutex);
	pp = Hash[Entrada];
	while ( (pp!=NULL) && (fin==0) ) {
		if (pp->valor == num)
			fin = 1;
		pp = pp->next;
	}
	release_lock(&mutex);
}