# Introducción a la memoria transaccional

## El problema de la concurrencia
Desde los primeros ordenadores multiproceso y multiusuario, la concurrencia ha
sido uno de los mayores problemas de la gestión a bajo nivel. Una situación
común generada por la concurrencia son las condiciones de carrera, en las que se
realizan operaciones simultáneas sobre recursos compartidos que pueden llegar a
dejar éstos en un estado inconsistente.

Para solucionar estos peligrosos conflictos, han surgido múltiples propuestas a
lo largo de la historia de la informática. Inicialmente estas soluciones han partido del software,
con la consecuente pérdida de rendimiento a cambio de fiabilidad. Sin embargo,
la tendencia en los últimos años es mover este problema al hardware, de forma
que podemos asegurar la fiabilidad minimizando la pérdida de rendimiento.

La solución que suele venir a la mente es la exclusión mutua. Con _locks_ o
_semáforos_ se definen secciones de código a las que denominamos **sección
crítica** que solo se ejecutarán si no se están ejecutando ya en otro núcleo del
procesador. De estarlo, se esperará hasta que se termine de ejecutar en otros
núcleos para entrar en la sección crítica.

Aunque esta estrategia es cada vez más eficiente por la inclusión de nuevas
instrucciones de hardware que simplifican y aceleran la implementación de este
tipo de estructuras, la exclusión mutua provoca una pérdida considerable de
rendimiento al evitar que ciertas secciones de código se puedan ejecutar
paralelamente. Por tanto, es trabajo del programador reducir al máximo el
tamaño de las secciones críticas para afectar lo menos posible al rendimiento.

## ¿Qué es la memoria transaccional?
El enfoque de memoria transaccional parte de la premisa contraria: en lugar de
partir de la existencia de posibles conflictos en las secciones críticas y por tanto
ejecutarlas siempre de forma secuencial, de forma que sea imposible que se acceda
a la vez a un mismo recurso compartido; en el caso de la memoria transaccional se
supone de forma optimista que la mayor parte de veces no habrá conflictos, pensando
en que dada la velocidad a la que se ejecutan las instrucciones en los procesadores,
el tiempo durante el cual se está accediendo a un recurso compartido es mínimo.

El reto que propone esta solución es detectar si ha ocurrido un conflicto y, en
caso de que se detecte, revertir los cambios realizados para pasar a ejecutar las
secciones críticas involucradas de forma segura. De esta forma, el coste inicial de
usar de forma segura un recurso compartido es menor, aunque potencialmente se traslada
ese coste al proceso de recuperación en caso de conflicto detectado.

La solución de memoria transaccional es conceptualmente simple pero implica
grandes problemas de rendimiento. Las primeras implementaciones de este modelo
fueron sobre software, y aunque el funcionamiento era satisfactorio, cumpliendo los requisitos de fiabilidad, el
rendimiento era muy pobre. <!-- TODO fuente -->. Sin embargo, ya se puede probar
este concepto sobre hardware en los procesadores Intel más recientes, con la promesa de sortear los problemas de rendimiento.

## Ventajas de la memoria transaccional
Este concepto permite mucha más flexibilidad para el programador. El ejemplo más
básico es el que usaremos en este trabajo: si es necesario tener una tabla hash
como recurso compartido, hasta ahora existía la opción de poner un cerrojo
a toda la tabla o poner un cerrojo a cada entrada de la tabla. La primera opción
facilita la programación a costa de empeorar el rendimiento, pues cualquier operación
sobre la tabla bloqueará el acceso a toda la tabla por parte de cualquier otro hilo de
ejecución, así que los accesos a cualquier entrada del recurso compartido serán
secuenciales. La segunda opción hace que los accesos a entradas distintas de la tabla
sean mucho más eficientes mientras no haya conflictos, caso en que habrá que esperar a
que termine la operación en esa entrada. Sin embargo, si usamos cerrojos, es necesario complicar el código
añadiendo gestión granular de cada entrada.

La promesa de la memoria transaccional une las ventajas de ambas opciones: por un
lado, la implementación de la zona crítica es tan sencilla como poner un cerrojo global
a toda la tabla, sin tener que gestionar cada entrada de la tabla con distintos cerrojos
En memoria transaccional sería tan sencillo como indicar toda la zona en la que puede
haber conflictos y el código se ejecutará concurrentemente mientras no los haya.

Con este trabajo quiero ver si, con el hardware actual a disposición de los consumidores,
concretamente un procesador Intel de última generación destinado a ordenadores domésticos, es factible crear
programas con un código sencillo en cuanto a la gestión de concurrencia pero con
el rendimiento de la gestión granular de locks. Quizás, la memoria transaccional
por hardware sea la manera de obtener lo mejor de los dos mundos.

## Historia de la memoria transaccional en hardware
Dado que las soluciones de memoria transaccional en software incurren en una
grave penalización de rendimiento para hacer el sistema seguro y reducir los
falsos positivos y falsos negativos, los fabricantes de procesadores han tratado
de implementar este paradigma directamente en el hardware.

Hasta recientemente, la solución por hardware no era viable para un uso
generalizado. Aunque aún no sabemos nada de cuándo será puesto a disposición del
público, en 2009 AMD anunció que estaban trabajando en unas extensiones para sus
procesadores denominadas _Advanced Synchronization Facility_ que prometían
llevar la memoria transaccional a los procesadores dirigidos a consumidores. Este fue el primer anuncio
de un proyecto de este tipo destinado a procesadores de uso común. Sin embargo, siete años después
seguimos sin oir noticias sobre este desarrollo. Por tanto, no podemos confirmar ni desmentir que
el proyecto esté acabado. De hecho, podriamos pensar que viendo que Intel apuesta por la memoria
transaccional sobre hardware, AMD se vea obligada a despertar el proyecto ASF para seguir compitiendo.

### Primeras apariciones de la memoria transaccional en hardware
Antes de AMD, Sun fue la primera empresa en prometer llevar la memoria
transaccional a un procesador de uso general. Durante el desarrollo de su procesador
llamado _Rock_, la comunidad estaba expectante ante las características que Sun
iba desgranando de su futurible producto. Además de ser el primer procesador
comercial con soporte para memoria transaccional, contaba con otras
características innovadoras como precarga del caché con los núcleos en desuso durante los
fallos de caché o ramas especulativas a nivel de hilo [@_rock_2016].

Como parte del
desarrollo de este procesador, se puso a disposición de desarrolladores externos
kits de desarrollo donde se pudo probar sobre hardware real la memoria
transaccional en la comunidad académica. También, Sun puso a disposición de los
desarrolladores un simulador de memoria transaccional por software para poder
hacer pruebas con las extensiones a la arquitectura SPARC que desarrollaron para
su procesador. Desafortunadamente, tras la compra de Sun por parte de Oracle en
2010, Sun confirmó que los planes para el desarrollo del procesador habían sido
cancelados.

Otros procesadores que soportan instrucciones de memoria transaccional desde hace años, aunque
con ciertas limitaciones, son el Blue Gene/Q de IBM o los procesadores RISC en algunas de sus
instrucciones. Sin embargo, estos procesadores no suelen estar dirigidos al público general
por su incompatibilidad con las instrucciones x86, sino que se enfocan a usos empresariales,
sistemas embebidos u otras aplicaciones más especializadas.

### Memoria transaccional en hardware con Intel
Tenemos que esperar hasta la cuarta generación de
procesadores Intel Core, denominados Haswell, para encontrar la primera comercialización de una
implementación de instrucciones de memoria transaccional sobre hardware destinado a usuarios
domésticos, que ocurrió en junio de 2013 por primera vez. Estos
procesadores incluían el conjunto de instrucciones TSX (Transactional
Synchronization Extensions) que permiten acceder a las capacidades de memoria
transaccional del procesador de dos maneras: de una forma más restringida pero
retrocompatible, con la interfaz HLE (Hardware Lock Ellision); o de una forma más
flexible pero incompatible con procesadores que no lo soporten con la interfaz
RTM (Restricted Transactional Memory). Más adelante elaboraremos sobre las diferencias
en estas dos interfaces.

Sin embargo, en agosto de 2014, Intel se vio obligado a desactivar las
funcionalidades de TSX [@wasson_errata] con una actualización de microcódigo en los
procesadores Haswell debido a un error que, bajo ciertas condiciones, provocaba
un comportamiento impredecible en el procesador. Este desafortunado evento
provocó que hubiera que esperar hasta la sexta generación de procesadores Intel
(Skylake) para poder usar estas funcionalidades con normalidad.

En la sexta generación de procesadores, la última lanzada por Intel, encontramos
las mismas extensiones que en la cuarta, cuando se hizo el primer lanzamiento.
La única diferencia es que a priori los errores encontrados en aquella han sido
resueltos. Las pruebas de este trabajo están realizadas sobre un procesador de
esta generación. A día de hoy no se
han reportado errores en la implementación de memoria transaccional de los procesadores
de sexta generacion.

## Funcionamiento de la memoria transaccional en hardware
Como ya adelantamos en la sección anterior, Intel proporciona un conjunto de instrucciones
denominado TSX que nos permite hacer uso de las capacidades de memoria transaccional en sus
procesadores de última generación; pero estas instrucciones pueden ser aprovechadas de dos maneras.
Estas dos formas coinciden con las dos implementaciones habituales en las que se suele dar la
memoria transaccional. En los siguientes apartados describiremos en más detalle cada uno
de estos acercamientos, acompañándolo de algunos ejemplos [@rajwar_going_2012].

### _Hardware Lock Ellision_ (HLE)
La más simple entre ambas interfaces para acceder a las capacidades de memoria transaccional
del procesador es la denominada elipsis de cierre de exclusión mutua
(_lock ellision_). Mientras que la programación con cierres de exclusión mutua
se considera pesimista al bloquear todos los accesos a la sección crítica para
evitar cualquier tipo de conflicto, este enfoque se basa en convertir la
programación con cierres de exclusión mutua en optimista, sin alterar sustancialmente el
código ya escrito con cerrojos de exclusion mutua. Para ello, el
procesador ejecuta inicialmente todas las secciones críticas como si no fueran
tales, de forma paralela transformándolas en transacciones; y en caso de que se produzca algún conflicto, el
procesador lo detectará y dará marcha atrás a las transacciones involucradas
para acto seguido volver a ejecutarlas secuencialmente pasando por los cierres
de exclusión mutua que ignoró en un principio.

En la implementación de Intel, las instrucciones del conjunto denominado HLE
(_hardware lock ellision_), que son las que implementan la elipsis de cierre de exclusión
mutua, son ignoradas sin provocar errores en procesadores que no las soportan,
por lo que automáticamente el código se ejecuta con todos los cierres de
exclusión mutua con seguridad y sin necesitar cambios para ser retrocompatible. Esta característica
permite actualizar fácilmente el código ya escrito con cerrojos, o escribir código que aproveche las
bondades de la memoria transaccional sin mucho esfuerzo y sin perder la compatibilidad con procesadores
anteriores a nivel de binario.

Una implementacion sencilla a nivel de código ensamblador sería la siguiente. En primer lugar, el código
que obtiene el cerrojo para comenzar la sección crítica:

```asm
        mov eax, 1
Try:    xacquire lock xchg mutex, eax
        cmp eax, 0
        jz Success
Spin:   pause
        cmp mutex, 1
        jz Spin
        jmp Try
```

Vemos que es un código bastante tradicional para adquirir un cerrojo. Algo que destaca es la
instruccion `xacquire`, que determina el comienzo de la zona transaccional con HLE. Como decíamos,
esta instruccion es en realidad una anotación que será ignorada por aquellos procesadores
anteriores que no la reconozcan, asegurando la retrocompatibilidad y la corrección del programa en
procesadores antiguos pues en dicho caso, ignorando la etiqueta se interpretará el cerrojo de
forma tradicional.

Este es el claro ejemplo de que realmente el coste de implementar un cerrojo normal o implementarlo
haciendo uso de HLE es el mismo, tanto en términos de productividad del programador como de compatibilidad
con procesadores.

Tras la sección crítica, el código para liberar el cerrojo sería el siguiente:

```asm
        xrelease mov mutex, 0
```

Siguiendo con la tendencia anterior, vemos que el único cambio es la anotación `xrelease` que, como
podríamos esperar, marca el fin de la zona transaccional. Con estos dos pequeños cambios tendremos
un programa que funciona correctamente y sin cambios en la práctica en un procesador no compatible
con TSX, pero que se ejecuta transaccionalmente a través de la técnica de la elisión de cierre de
exclusión mutua sin ningún esfuerzo para el programador. Esto implica que el código de la sección
crítica se ejecutará en paralelo mientras no haya conflictos, y en el caso de que los haya (u ocurra
cualquier otro problema que aborte la transacción) será el propio procesador quien deshaga los cambios
realizados hasta el momento y repetirá la ejecución usando el cerrojo tradicional. Todo esto, sin ningún
tipo de runtime y aprovechando la eficiencia de la implementación en hardware.

### _Restricted Transactional Memory_ (RTM)
Frente a la propuesta más conservadora que plantean las instrucciones HLE, Intel también proporciona
dentro del conjunto TSX una segunda interfaz que pierde la retrocompatibilidad con anteriores procesadores
a cambio de aportar una mayor flexibilidad a la resolucion de los conflictos y otros tipos de errores que
se pueden dar en las transacciones, obteniendo a cambio un rendimiento potencialmente mejor al permitir
al programador la posibilidad
de adaptar la estrategia de respuesta frente a errores a cada carga de trabajo concreta.

El conjunto de instrucciones RTM (_Restricted Transactional Memory_) permite
designar una parte del código como transaccional con las instrucciones
`XBEGIN` y `XEND`, incluso de forma anidada; y permite indicar un desplazamiento
hasta una instrucción en la que se implemente un camino seguro alternativo, o
incluso una lógica distinta que, por ejemplo, decida si reintentar la transacción
en función de la información proporcionada por el procesador.

Un ejemplo útil de implementacion de RTM es el benchmark que desarrollaremos para caracterizar
el comportamiento de las instrucciones TSX en posteriores secciones. Pese a la sencillez de uso de HTM,
esa estrategia no sirve para caracterizar el comportamiento pues no nos permite acceder a las razones
del fallo de una transacción, a diferencia de las RTM.

Veremos también un ejemplo sencillo de implementación de RTM en ensamblador. Comenzamos con el código
que realiza la adquisición del cerrojo:

```asm
Retry:  xbegin Abort
        cmp mutex, 0
        jz Success
        xabort $0xff
Abort:
        ; Comprobar el registro EAX para conocer la razón del fallo.
        ; Aquí podemos elegir si reintentar la transaccion, realizar
        ; la operación con un lock tradicional...
```

El código comienza con la instrucción `xbegin`, que recibe como parámetro la etiqueta
del bloque de código que gestionará los fallos de la transacción, donde se podrá
leer la razón del fallo y decidir cómo continuar: repitiendo la transacción,
adquiriendo un cerrojo tradicional u otras alternativas más dependientes de cada
problema que se puedan adaptar mejor. De aquí es de donde surge la flexibilidad de RTM:
es nuestro propio código quien gestiona los fallos en las transacciones, con lo que podemos
resolverlos con total libertad. Por supuesto, esto conlleva algo más de trabajo por parte del
programador si lo comparamos con la implementación de HLE sobre un programa ya existente, pero
vemos que el codigo no es mucho más complejo que el de un cerrojo tradicional.

La adquisición del cerrojo salta a la zona crítica si el cerrojo no ha sido adquirido aún
(podria haber hilos que lo cojan directamente sin transacciones, o transacciones fallidas
que se hayan resuelto usando un lock tradicional), y de lo contrario vemos una instrucción nueva,
`xabort`, que fuerza el fallo de la transacción. `xabort` recibe como parámetro un valor que
se pasará como código de error al código que gestiona los fallos [@intel_intel_ISE].

Tras la sección crítica, podríamos ver un código como el siguiente:

```asm
        cmp mutex, 0
        jnz release_lock ; Salto a "mov mutex, 0"
        xend
```

Este código puede gestionar simplemente el fin de la transacción, que se marca con
la instrucción `xend`, o bien la liberación del cerrojo de exclusión mutua.

## Implementación en hardware de las instrucciones para memoria transaccional hardware de Intel (TSX)
Los dos sistemas anteriormente descritos (HLE y RTM) tienen interfaces distintas
que habilitan posibilidades distintas. HLE permite convertir a código transaccional
un código habitual con cierres de exclusión mutua, manteniendo la retrocompatibilidad;
mientras que RTM permite soluciones de memoria transaccional más creativas y posiblemente
más adaptadas a cada problema y por tanto más eficientes, a cambio de que el software
desarrollado con RTM no pueda ser ejecutado en procesadores que no dispongan de las
extensiones TSX.

A pesar de tener estas dos interfaces distintas, el funcionamiento interno de ambas es el mismo.
Paso a describir su comportamiento desde diferentes propiedades.

### Detección de colisiones
Intel se decantó por implantar memoria transaccional con aislamiento fuerte (_strong isolation_).
Esta forma de memoria transaccional es más segura que otras alternativas como el aislamiento débil
(_weak isolation_). Aunque no entraremos en el debate entre ambas propuestas, que
cuentan con entusiastas defensores y detractores, sí describiremos en qué se traduce
esta decisión a la hora de desarrollar código transaccional con TSX.

La principal diferencia entre ambos acercamientos es que un aislamiento fuerte no solo
detectará colisiones entre secciones de código transaccional, sino que también detectará
situaciones en las que código no transaccional afecta a la integridad de una transacción.

El desarrollo de sistemas de memoria transaccional ha estado limitado más allá de la
investigación a la memoria transaccional por software que, como hemos comentado,
incurre en unas graves penalizaciones de rendimiento. Por tanto, aunque algunos
investigadores entienden que el aislamiento fuerte es una propiedad deseable, no se
ha podido aplicar este método puesto que limita más el rendimiento que el aislamiento débil [@minh_effective_2007].
Por tanto, las soluciones de memoria transaccional software no se pueden permitir
aislamiento fuerte. Al implementar la memoria transaccional en hardware, podemos
disfrutar de la seguridad del aislamiento fuerte sin perder demasiado rendimiento.

Para implementar este aislamiento fuerte y detectar colisiones en los accesos y escrituras
a la memoria que puedan invalidar la transacción, ya sean desde otros bloques
transaccionales o desde código no transaccional, la implementación de Intel
hace un seguimiento de las líneas de caché que se leen y se escriben en cada
transacción (_read-set_ y _write-set_). Actualmente, estas líneas de caché tienen un
tamaño de 64B.

Al conocer las líneas de caché leídas y escritas, el procesador puede detectar cuándo <!--alguna
de las líneas que la zona transaccional ha leído es escrita, o cuándo alguna de las líneas
que ha escrito es leída.--> se produce algún conflicto, ya sea entre secciones de código
transaccional o no, que obligue a abortar la transacción en curso.

La elección conservadora de optar por el aislamiento fuerte evita posibles condiciones de
carrera, conflictos y otros tipos de errores que podrían no ser detectados con un
aislamiento débil, pero es un lujo que podemos permitirnos al estar el sistema
implementado en silicio.

### Anidamiento de secciones transaccionales
Tanto HLE como RTM permiten anidar transacciones, pero ambos métodos tratan estos anidamientos
de la misma manera. El procesador tiene una constante cuyo valor depende de la implementacion
y no es accesible al sofware que define el máximo número de transacciones anidadas permitidas
para cada tipo de interfaz (HLE o RTM), así como un contador que tampoco es accesible
dinámicamente que suma uno al abrir una zona transaccional y resta uno al cerrarla. Cuando el
valor del contador alcanza el valor prefijado en la constante, todas las transacciones
que estén anidadas en ese momento fallarán por exceder el límite de anidamiento.

Hay que tener en cuenta que el procesador tratará todas las transacciones anidadas como una
gran transacción monolítica, lo que explica que todas las transacciones no solo serán abortadas
si se supera el límite de anidamiento, sino que en el caso de que cualquiera de las
transacciones internas falle por cualquier motivo, todas las transacciones anidadas fallarán
también.

### Instrucciones que invalidan las transacciones
La implementación de memoria transaccional de Intel hace que ciertas instrucciones del procesador
dentro de una transacción provoquen que sea abortada. La especificación de las instrucciones TSX no
define qué instrucciones provocarán que la transacción se anule, puesto que el objetivo
es que cada vez menos instrucciones provoquen estos fallos.

Algunas de las instrucciones que obligan a abortar una transacción son:

- Funciones que impriman a la salida estándar (`printf`). <!-- ¿Seguro? -->
- La instrucción `PAUSE`, usada para hacer eficientes los bucles de espera activa
que se suelen usar con los cerrojos de bloqueo mutuo.
- La instrucción `CPUID`, que permite obtener información sobre la versión y
capacidades del procesador en tiempo de ejecución.
- Operaciones del coprocesador de coma flotante, del subconjunto de instruccciones `x87`.
- Operaciones `MMX`, que operan sobre los registros del coprocesador de coma flotante.
- Operaciones simultáneas en una misma transacción sobre los registros `XMM`, que son
usados para operaciones de coma flotante en las instrucciones `SSE` (_Streaming SIMD
Extensions_), y los registros `YMM`, usados por las `AVX` (_Advanced Vector Extensions_) que
expanden el tamaño de los registros SIMD hasta 256 bits.
- Cambios a cualquier campo que no sea de estado en los registros tipo `FLAGS`.
- Interrupciones.
- Eventos de E/S.

Otro evento habitual que cancela una transaccion es un cambio de contexto, puesto
que en esa situación se anularán la mayoría de las líneas de caché del procesador.
Esto hace que sea arriesgado realizar operaciones del sistema operativo, puesto que
al ceder el control a éste, se producirá un cambio de contexto con las consecuencias
anteriormente mencionadas.
