# Análisis y diseño de una aplicación sintética para caracterizar la memoria transaccional

## Objetivos de la aplicacion sintética
Para estudiar el comportamiento de la memoria transaccional, caracterizarlo y ver para qué
tipo de aplicaciones es éste acercamiento más apropiado, haremos uso de un benchmark sintético
con la intención de simular a través de unos parámetros las diversas condiciones que se pueden
dar en el acceso a una base de datos que se encuentre tras un servidor web, tratando de dar
respuesta a las preguntas que nos hacíamos en la introducción.

En las siguientes subsecciones describiremos pormenorizadamente cada uno de los objetivos que nos
proponemos para este benchmark sintético. Tras la descripción, analizaremos qué características
impondrán al diseño final del benchmark para que el resultado final pueda cumplir estos objetivos.

### Simular el modelo de una base de datos
Partiendo de la problemática que describimos en el
trabajo, aspiramos a hacer un benchmark que, sabiendo que no puede equipararse al uso de una base
de datos en producción, sí nos pueda dar nociones sobre en qué situaciones se puede comenzar a
utilizar esta incipiente tecnología. Por tanto, deberemos intentar que el benchmark sea comparable
al funcionamiento de una base de datos en memoria.

Para simular esta estructura de datos, usaremos una tabla hash. La misma estará compuesta con un
vector en el que cada uno de sus elementos es la cabeza de una lista enlazada. Podríamos comparar
esta estructura, de forma simplificada, a la de una tabla de una base de datos moderna cacheada
en la memoria para un acceso rápido. La memoria transaccional o el cerrojo de exclusión mutua
nos proporcionarán las garantías de que los datos se mantienen consistentes. La consistencia a
través de la red en múltiples nodos queda fuera del ámbito de este trabajo.

Además, no solo la forma de la estructura de datos deberá asemejarse a la que podría usar una base
de datos en memoria, sino que también los patrones de acceso deberán ser similares. Como hemos
puesto el foco en simular una base de datos que da servicio a un concurrido servidor web, supondremos
que se lanzan varios hilos de forma concurrente, con algunos de ellos de lectura y otros de escritura.
Para simular la naturaleza estocástica de los accesos a un servidor web, tanto las lecturas como las
escrituras se harán a posiciones aleatorias de la tabla. Además, la tabla será previamente rellenada
con datos aleatorios.

### Altamente parametrizable
Debemos hacer un análisis de los parámetros que pueden influir en
el rendimiento de una ejecución con memoria transaccional de forma que podamos automatizar a
través un _script_ la ejecución de este benchmark con todas las posibilidades dentro de los rangos
de parámetros que decidamos. Esto permitirá facilitar las pruebas, ya que además de permitir una
automatización sencilla de la ejecución, evitamos nuevas compilaciones para cambiar los parámetros
y agilizaremos la investigación al poder partir de una ejecución amplia para luego dirigir el foco
a las partes que consideremos más interesantes.

Es importante en este punto decidir qué variables es interesante parametrizar para poder estudiar
fielmente situaciones que puedan caracterizar el comportamiento de las extensiones TSX. Para ello
no podemos olvidar el anterior capítulo de este trabajo en el que analizamos a nivel técnico esta
tecnología. De ahí deducimos varios parámetros que pueden cambiar seriamente el comportamiento de
las transacciones.

En el punto anterior ya hemos hablado de la estructura de datos elegida para representar la base
de datos en memoria, con lo que de aquí en adelante hablaremos en función de dicha estructura.

Los primeros parámetros que necesitamos manipular serán las dimensiones de la tabla. Éstas afectarán
a la probabilidad de que las transacciones aborten, y en función de qué dimensión se modifique,
variará más notablemente un tipo concreto de fallos u otro. Extenderemos sobre este punto en el siguiente
apartado.

Otro parámetro que deberíamos poder modificar para caracterizar el comportamiento de la
memoria transaccional es el número de hilos que están trabajando simultáneamente sobre la estructura.
Aquí es donde podremos ver si realmente la memoria transaccional puede solucionar el cuello de botella
que impone un cerrojo de bloqueo mutuo a toda la estructura, si realmente se puede superar el mismo con
la memoria transaccional por hardware con una dificultad de desarrollo similar, y en qué condiciones.

Por supuesto, también es necesario poder parametrizar la carga de trabajo de la aplicación. Deberíamos poder
especificar el número de accesos, ya sean de lectura o escritura, que realiza cada hilo.

### Caracterización en base a los tipos de fallos
Como nuestro objetivo es caracterizar el comportamiento de la memoria transaccional, nos es necesario
conocer cómo afectan los cambios en los parámetros a dicho comportamiento. Nos valdremos para ello por
una parte de la comparativa en el tiempo de ejecución frente a los cerrojos tradicionales, pero también
nos será útil conocer la razón de los posibles fallos en las transacciones para entender cómo afectan al
rendimiento.

En primer lugar, nos fijamos en los fallos por capacidad. Este tipo de fallos ocurren cuando antes
de que termine una transacción, el procesador ha tenido que invalidar una línea de caché que estaba
siendo editada en una zona transaccional debido a que no hay más espacio en la caché y, según la
política de gestión de la caché del procesador, esa línea tenga que ser invalidada. En la implementación actual
de las extensiones TSX no se da ninguna prioridad especial a las líneas de caché usadas en una transacción así
que éstas pueden ser invalidadas con la misma facilidad que el resto de líneas.

El parámetro que más efecto
tiene en los fallos por capacidad en nuestra simulación de base de datos es la longitud de la lista enlazada,
pues cuanto más larga sea la misma, más se necesitará llenar la caché para alcanzar el dato deseado. Aunque
el acceso es aleatorio, la probabilidad de navegar por un número mayor de nodos de la lista enlazada, llenando
la caché, será mayor cuanto mayor sea la longitud de la lista enlazada.

Otro tipo de fallos que pueden afectar seriamente al comportamiento de la memoria transaccional son
los fallos por conflicto. Estos ocurren cuando se detecta un conflicto, es decir, cuando se producen
lecturas y escrituras, o varias escrituras, durante una transacción sobre una misma línea de caché.
Recordemos que al ser éste un modelo de memoria transaccional con aislamiento fuerte, incluso un
acceso desde fuera de cualquier código transaccional dispararía un fallo por conflicto abortando la
transacción; sin embargo todos los accesos a la estructura compartida se harán dentro de transacciones
en este caso.

Para ver el efecto de este tipo de fallos, usaremos como parámetro el tamaño del vector que
contiene las cabezas de las listas enlazadas. Cuanto mayor sea este tamaño, más probable es que dos hilos
caigan (como ya hemos dicho, aleatoriamente) en listas enlazadas distintas, y por tanto no compartan
líneas de caché para acceder a los datos. A su vez, reduciendo el tamaño del vector podremos forzar conflictos.

### Sencillo en el diseño y la implementación
Puesto que de un benchmark se trata, éste software no es el objetivo final del trabajo, sino
los datos que podamos sacar de él y la información que podamos deducir de ellos. Sin descuidar
su diseño e implementación, debemos tener claro que debe primar la sencillez que nos permita
iterar con rapidez y evitar posibles errores.

Una decisión que nos impone la técnica es el uso de un lenguaje de bajo nivel como ensamblador, C o C++. 
Para facilitar la tarea del desarrollo, evitaremos el lenguaje ensamblador. El desarrollo en C o C++ 
no dificulta la tarea puesto que a través del compilador GCC [@_x86_gcc],
tenemos un soporte asequible para usar las extensiones TSX sin tener que bajar al nivel del código
ensamblador.

Para acelerar la implementación y la iteración, usaremos un paradigma de programación imperativa,
usando funciones para facilitar la lectura y almacenando globalmente el estado de la aplicación
cuando sea necesario. Intentaremos limitar el código tanto en número de ficheros como en tamaño de
los mismos, aspirando a que el código sea lo suficientemente sencillo como para poder describirlo
con claridad solamente con un fichero de código y otro de encabezado.

También, para simplificar la implementación de la memoria transaccional, usaremos las instrucciones
RTM usando un cerrojo de exclusión mutua global como plan b. Como decíamos anteriormente, las
instrucciones RTM permiten una mayor flexibilidad en cuanto al plan b en caso de fallo de la transacción,
por lo que es probable que los resultados que obtengamos no sean óptimos comparándolos con la flexibilidad
que ganaríamos aprovechándonos de la información sobre el fallo que nos da el procesador y de alguna
solución adaptativa que optimice para cada situación. Esta investigación sobre los límites de RTM queda
fuera del alcance de este trabajo.

### Comparar con la ejecución tradicional con cerrojos
Para maximizar la utilidad de este estudio, buscaremos que el propio benchmark nos proporcione datos comparativos
entre la ejecución usando TSX, y exactamente la misma ejecución pero usando un cerrojo tradicional. Para que
la comparación sea justa, ambos métodos emplearán una restricción a la estructura de datos completa,
no por líneas. De esta forma podremos comparar en qué casos es más eficiente el uso de memoria transaccional
aún cuando la granularidad de las restricciones y el esfuerzo a la hora de programarlo es comparable. Es obvio
que debemos garantizar que éstas ejecuciones no son lanzadas concurrentemente para que ambas puedan utilizar
todo el potencial de la máquina en igualdad de condiciones.

## Análisis y diseño de la aplicacion de benchmark sintético
En función de estos objetivos, plantearemos las consecuencias que tendrán éstos en el diseño final de
la aplicación.

### Interfaz
Tanto por simplicidad en el desarrollo como por simplicidad en el uso, la interfaz ideal para esta situación
es la línea de comandos. Debería ser suficiente con llamar al ejecutable indicando como parámetros todos los
valores variables que hemos indicado anteriormente. Una interfaz de línea de comandos podria tener la siguiente
forma, suponiendo que los valores con la forma `<valor>` son requeridos y aquellos con la forma `[valor]` son
opcionales.

```
./benchmark <Número de hilos> <Número de iteraciones>
        <Número de entradas en la lista enlazada> <Tamaño de la tabla>
        <Fichero de salida> [Verboso]
```

Para facilitar el posterior tratamiento de datos, sería bueno obtener éstos en un formato sencillo pero estándar
de forma que pueda ser usado sin problema por sean las que sean las herramientas que se escojan para este fin.
El formato CSV reune satisfactoriamente estas condiciones.

### Estructura interna
Internamente, el programa constará de un fichero de encabezado `TablaHash.h` con las definiciones de todas las
variables y funciones globales. Ese fichero será usado desde el fichero con el código del programa, `TablaHash.c`.
En él tendremos una función `main()` que irá ejecutando secuencialmente las operaciones indicadas. Concretamente,
esta función main seguiría el siguiente esquema:

- Leer y almacenar los parámetros.
- Inicializar la tabla según los parámetros dados, rellenándola con valores aleatorios.
- Hacer una ejecucion parametrizada lanzando los hilos indicados con la carga de trabajo indicada usando un cerrojo
global de exclusión mutua. Tomar tiempos y almacenarlos.
- Repetir la misma ejecución pero usando memoria transaccional en lugar del cerrojo, midiendo de nuevo el tiempo.
- Obtener las estadísticas de fallos de transacciones.
- Con los tiempos y las estadísticas de fallos, escribir una línea en formato CSV al fichero que se indicó como
parámetro indicando los parámetros de esta ejecución y todos los datos recogidos.
- Tras esto, el programa finaliza indicando al script que lo lanzará que ya puede ejecutarlo con la siguiente variación
de los parámetros.

Las funcionalidades que se llaman desde la función `main()` se dividirán en funciones para aumentar la legibilidad.
Detectamos las siguientes partes que se pueden dividir en funciones:

- La creación de la tabla hash global.
- El rellenado de la tabla con datos aleatorios.
- La adquisicion y liberación del cerrojo global o, en su caso, la apertura y cierre de la zona transaccional.
- La creación de los hilos de edición y búsqueda.
- Las funciones que ejecutarán dichos hilos, que repetirán su acción en bucle para simular una carga de trabajo.
- Las funciones que hacen la edicion y la búsqueda en sí, que serán ejecutadas por los hilos.
- La espera por todos los hilos, para saber cuándo todos ellos han terminado.
- La obtencion de las estadísticas de fallos de transacciones tras la ejecución.

La ejecución transaccional conlleva un problema adicional, y es que debemos compaginar el uso de la memoria
transaccional con el del cerrojo global que se usará si la transacción aborta. La sección de código que abre la zona
crítica será igual para todas las transacciones, pero la que la cierra deberá por un lado volver a ejecutar la
transaccion con un cerrojo si ésta ha fallado, y por otro lado comprobar si se está usando o no el cerrojo para
liberarlo o bien llamar a la función que cierra la sección transaccional. Para conseguir esto, nos vemos obligados
a comprobar si tenemos el cerrojo y por tanto debemos liberarlo al terminar; podemos hacer uso de una comparacion como
esta, siendo `mutex` el cerrojo global:

```c
if ((*mutex).__data.__lock == 0) {
    // Estamos en transaccional
} else {
    // Tenemos el cerrojo y hay que liberarlo
}
```

Es probable que el compilador se queje del acceso a este campo privado, pero será un mal menor.

### Camino transaccional y camino seguro
En la ejecución transaccional tenemos a priori dos opciones que nos proporciona el conjunto TSX. Bien podemos
usar HLE, con la ventaja de la sencilla implementación, pues se basa en cerrojos tradicionales añadiendo una simple
anotación en la adquisición y liberación del mismo para convertir la ejecución en transaccional; bien podemos
usar RTM, que a cambio de una implementación algo más compleja, tenemos más flexibilidad sobre qué hacer
tras el fallo de una transacción. Sin embargo, hay un factor a mayores que decanta la decisón por RTM, y es
que éste es el único conjunto de instrucciones que permite recopilar estadísticas sobre las razones de fallo
de las transacciones, un hecho vital para una aplicación como ésta que trata de caracterizar el comportamiento
de la memoria transaccional, no limitándose únicamente al rendimiento.

Sin embargo, en nuestro caso nos sería suficiente con el comportamiento de HLE: ejecutar de forma transaccional y,
en caso de fallo, deshacer los cambios para pasar a ejecutar con un cerrojo tradicional que asegure la
corrección de la ejecución. Con RTM podemos generar este comportamiento sin perder la posibilidad de recopilar
las valiosas estadísticas sobre los motivos de fallo.

Por tanto, a la hora de desarrollar la aplicación deberemos crear un código de adquisición y liberación del
cerrojo de exclusión mutua que compatibilice el uso de cerrojo tradicional con la gestión de las transacciones
usando RTM por dos razones: por una parte, como hemos dicho, haremos que el código transaccional revierta
a un cerrojo cuando la transacción falle, asegurando la corrección del programa; y por otra parte, la ejecución
con la que compararemos la ejecución transaccional es una ejecución que usa cerrojos normales. Si nuestro
programa tiene funciones para gestionar ambos casos, podemos manejar ambas ejecuciones (transaccional y con
cerrojos tradicionales) usando el mismo código.

La función que adquiere el cerrojo o marca el inicio de la transacción podría tener la
siguiente forma:

```c
void acquire_lock(pthread_mutex_t *mutex, int use_tsx) {
	unsigned int status;

	if (use_tsx) {
		if ((status = _xbegin()) == _XBEGIN_STARTED) {
            // Transacción iniciada.
			if ((*mutex).__data.__lock == 1) {
                // Si, a pesar de estar en transacción, alguien ha
                // adquirido el cerrojo, la transacción termina y
                // esperamos para cogerlo.
				_xend();
				pthread_mutex_lock(mutex);
			}
            // Si no, la transacción continúa sin adquirir el cerrojo.
		} else {
            // La transacción falló.
			// Recogida de estadísticas usando la variable "status".
			pthread_mutex_lock(mutex);
		}
	} else {
		pthread_mutex_lock(mutex);
	}
}
```

La función que libera el cerrojo o termina la transacción podría ser de la siguiente
forma:

```c
void release_lock(pthread_mutex_t *mutex, int use_tsx) {
	if (use_tsx) {
        // Si la ejecución es transaccional...
		if ((*mutex).__data.__lock == 0) {
            // y no se tuvo que coger el cerrojo, la transacción
            // termina con éxito.
			_xend();
		} else {
            // De lo contrario, la transacción ha fallado y se libera
            // el cerrojo que se adquirió.
			pthread_mutex_unlock(mutex);
		}
	} else {
        // Si la ejecución no es transaccional, simplemente se libera el cerrojo.
		pthread_mutex_unlock(mutex);
	}
}
```

### Modelo de paralelismo con hilos
Ya hemos hablado de que la aplicación sigue un modelo de paralelismo a través de hilos tradicionales,
un modelo que la memoria transacional nos permite paralelizar. La ejecución con cerrojos de exclusión
mutua sigue un modelo similar, sin olvidar que en el momento en que se necesite acceder al recurso compartido,
la ejecución se serializaría. Con la memoria transaccional podemos evitar este problema sin perder la corrección
del programa.

En el siguiente diagrama se describe este modelo que, como decíamos, simula un servidor de alto rendimiento que
tiene que servir lecturas y escrituras de datos aleatorias desde una tabla compartida.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=\textwidth]{img/diagrams/thread-model.png}
    \end{center}
    \caption{Diagrama que describe el modelo de paralelismo con hilos del benchmark sintético}
\end{figure}

En el diagrama vemos que un número indeterminado de usuarios realizan peticiones que llegan hasta el hilo principal
de nuestra aplicación sintética. En nuestra ejecución, estos usuarios se modelan como iteraciones en las que se realiza
alguna tarea. Después, el hilo principal es el encargado de lanzar los hilos que realizan las tareas. En nuestro modelo,
dividimos a la mitad entre escrituras y lecturas. Todos estos hilos acceden simultáneamente a una estructura compartida
que almacena la información sobre la que se trabaja. En nuestro caso, es una tabla hash.

## Otras herramientas
Además del propio benchmark, será necesario crear otras utilidades auxiliares para el procesado de los datos generados
por éste. Para ello optamos por usar Python, dado que crearemos scripts ad-hoc para cada gráfica que deseemos generar
y necesitamos la flexibilidad y agilidad que proporciona este lenguaje. Trataremos de reutilizar en la medida de lo
posible los algoritmos que creemos para el procesado de los CSV dividiéndolo en varias etapas (procesado inicial,
transformación, generación de graficas). El estado final de estos scripts se puede encontrar en el código anexo al
trabajo.

## Configuración del entorno
Para el desarrollo de este trabajo es necesario contar con una máquina que disponga de procesador Intel suficientemente
reciente como para ejecutar las instrucciones necesarias de la extensión TSX. La Escuela puso a mi disposición una
máquina que cumple dicho requisito y que tiene las siguientes características [@_intel] [@cutress_intel] [@martinez_intel] [@_core]:

- Procesador:
    - Fecha de lanzamiento: 2 de junio de 2015
    - Modelo: Intel(R) Core(TM) i7-5775C CPU @ 3.30GHz
    - Tamaño caché: 6 MB
    - Núcleos: 4
    - Máxima potencia permitida: 65W
    - Frecuencia DRAM: 1600MHz
    - Hilos de ejecución: 8 (Hyper-Threading, 2 procesos por núcleo físico)
    - Arquitectura: x86_64
    - Litografía: 14 nm
    - Caché L1 datos: 4x32KB (8-way)
    - Caché L1 instrucciones: 4x32KB (8-way)
    - Caché L2: 4x256KB (8-way)
    - Caché L3: 6MB (12-way)
    - Caché L4: 128MB
    - Tamaño de línea de caché: 64 bytes
- GCC: `gcc (Ubuntu 4.8.4-2ubuntu1~14.04.1) 4.8.4`
- Memoria RAM: DDR3 4GB
- Kernel: `Linux version 3.19.0-25-generic (buildd@lgw01-20) (gcc version 4.8.2 (Ubuntu 4.8.2-19ubuntu1) ) #26~14.04.1-Ubuntu SMP Fri Jul 24 21:16:20 UTC 2015`

Es interesante fijarse en el primer nivel de caché, pues es en dicho nivel en el que el procesador lleva el conteo de líneas de caché leídas y editadas durante
transacciones para detectar los conflictos [@rajwar_going_2012]. Vemos que tenemos en total 256KB de memoria caché de primer nivel,
siendo la mitad para datos y la otra mitad para instrucciones.

Como veremos, esta limitación afecta al rendimiento de las ejecuciones con memoria transaccional. También afectará,
por tanto, el tamaño de línea de caché que en este modelo está limitado a 64 bytes. Además, el caché tiene una asociatividad de ocho, que teniendo en cuenta el _multithreading_ en
procesadores Intel como éste que permite que haya más de un hilo compartiendo un mismo núcleo del procesador, hace que
la asociatividad de la caché pueda ser cercana a cuatro en la práctica.
