# Modelado de memoria transaccional hardware con una aplicación sintética
Este trabajo comenzó con una introducción en la que se exponían las motivaciones del mismo. A continuación,
vimos la introducción al concepto de memoria transaccional para gestionar la concurrencia, concretando en la memoria
transaccional por hardware y la implementación de Intel en sus procesadores de nueva generación de
este concepto a través de las extensiones TSX. Justo después continuamos con el análisis y diseño de la aplicación de benchmark
sintético necesaria para, usando estas extensiones TSX, cumplir con los objetivos del trabajo.

Ahora, el paso lógico es entrar en la sección en que usaremos la aplicación creada para modelar y caracterizar el
comportamiento de este acercamiento para gestionar el acceso concurrente a recursos compartidos
evitando bloqueos en la medida de lo posible sin poner en juego la corrección de los programas: la memoria
transaccional a través de las instrucciones TSX en procesadores Intel.

## Descripción de la aplicación de benchmark sintético creada
Se ha realizado una aplicación sintética diseñada específicamente para probar el comportamiento
de las transacciones frente a una aproximación más tradiccional con cerrojos de exclusión mutua,
y ver en qué condiciones merece la pena usar memoria transacciona hardware en lugar de cerrojos
bajo la implementación que Intel ha puesto a disposición de los consumidores.

### Simulando una base de datos
Esta aplicación, que denominaremos `TablaHash`, emula una tabla a la que se accede aleatoriamente,
tratando de simular el comportamiento que puede tener una base de datos en un servidor web u otras
aplicaciones de gran carga concurrente. Como se expuso anteriormente, entendemos que éste puede ser
un buen caso de uso para la memoria transaccional.

Esta tabla está compuesta de dos dimensiones. Cada elemento
de la tabla contiene un valor y forma parte de una lista enlazada. El número de listas enlazadas que
forman la tabla lo denominaremos **tamaño** de la tabla, mientras que la cantidad de
elementos que forman cada lista enlazada será el **número de entradas** de la lista. Nos referiremos
a estas dimensiones con los nombres anteriores durante el resto del trabajo.

Lo primero que hace la aplicación en cada ejecución es rellenar esta tabla con valores aleatorios. Con
esta aleatoreidad pretendemos evitar la secuencialidad en los accesos a caché y así simular patrones de
acceso aleatorios como los que podemos ver en servidores web de alto rendimiento.

Después, lanzará un número de hilos de búsqueda y escritura de valores en esta tabla de forma
simultánea, con el objetivo de estudiar el comportamiento en una situación de conflictos entre
lecturas y escrituras. Este proceso se repetirá dos veces: primero usando un cerrojo de exclusión
mutua sobre la tabla, y después usando la memoria transaccional hardware que nos proporciona el procesador.

### Comparativa de la ejecución con cerrojos tradicionales y con memoria transaccional
La implementación con cerrojos de exclusión mutua es muy tradicional: cuando el hilo va a acceder
a la sección crítica, que en este caso son los accesos a la tabla, pide el cerrojo. Si nadie lo tiene en ese momento,
lo adquiere y entra en la sección crítica bloqueando el acceso al resto de hilos para evitar cualquier
tipo de conflicto. Tras realizar su tarea en el recurso compartido, el hilo sale de la sección crítica
liberando el cerrojo y permitiendo que uno de los demás hilos pueda adquirirlo para acceder a la sección
crítica. Nada nuevo.

En cambio, la implementación que aprovecha la memoria transaccional hardware no bloquea el acceso al resto
de hilos al entrar en la sección crítica. Aunque el código es el mismo para los hilos, las funciones que gestionan
la adquisición y liberación de los cerrojos pueden usar directamente un cerrojo de exclusión mutua tradicional,
que es lo que hace en el caso anterior; o bien puede iniciar una transacción con las instrucciones de la extensión
TSX, concretamente las RTM en nuestro benchmark.

En caso de que todo vaya bien, tendríamos potencialmente varios hilos accediendo a un
mismo recurso compartido con la garantía de que no se han producido conflictos, ya sea porque todos estaban
leyendo, porque se leyeron líneas de caché distintas a las que se editaron, o porque nadie editó a la vez elementos
en la misma línea de caché. La clara ventaja de este acercamiento es que esta optimización de la detección de
conflictos ha sido totalmente transparente para nosotros, ya que confiamos en el propio procesador para realizarla
de manera óptima, y además se ejecuta directamente sobre hardware. El coste a la hora de programarlo ha sido el mismo que al configurar un cerrojo de bloqueo mutuo
global.

Sin embargo, si la transacción va mal y resulta ser abortada, el procesador anulará todos los cambios que haya
realizado nuestro hilo de forma transparente para el resto de hilos, sin que nadie pueda llegar a ver los cambios.
Además, podremos ver el motivo del fallo y recuperarnos de él. En nuestro caso, aunque usamos las instrucciones
RTM que permiten soluciones más creativas y adaptadas al problema y al tipo de fallo de la transacción, la solución
que aplicamos es repetir la operación usando un cerrojo de exclusión mutua tradicional, que nos da garantía de que
la operación se realizará correctamente. Esto no es problema pues el modelo de memoria transaccional de Intel para
TSX, al ser de aislamiento fuerte, detecta los conflictos en los datos, vengan o no de código transaccional. Usamos
las instrucciones RTM debido a que nos permiten obtener estadísticas sobre la razón de los fallos, necesarias para
la caracterización.

## Parámetros de la aplicación
La aplicación está diseñada para la realización de los benchmarks, con lo que cuenta con cinco parámetros que se
pueden configurar para probarla en distintas condiciones:

- Número de hilos: Define la cantidad de hilos de lectura y de escritura que se pondrán a trabajar contra la
estructura compartida.
- Número de iteraciones: Define la cantidad de operaciones que hará cada hilo antes de terminar.
- Número de entradas: Define la cantidad de elementos que forman cada lista enlazada de la tabla.
- Tamaño de la tabla: Define la cantidad de listas enlazadas que formarán la tabla.

Además de estos parámetros, se debe indicar un nombre de fichero donde el programa, tras su ejecución, escribirá
una línea en formato CSV indicando los parámetros con los que se le ha llamado, el número de transacciones que
se realizaron correctamente, el número de transacciones que fallaron, la cantidad de transacciones que han fallado
por cada tipo de fallo posible, el tiempo que tardó la ejecución con cerrojos de exclusión mutua, y el tiempo haciendo
uso de la memoria transaccional. Por último, se le puede indicar al programa que sea verboso: en ese caso mostrará
algunos detalles sobre el tipo de hilos que ejecuta y mostrará más información intermedia durante el proceso por la
salida estándar. En caso de un error en la introducción de los parámetros, el programa imprimirá cuáles son éstos para
recordarlo.

Es interesante recordar en este punto el análisis realizado en el capítulo anterior en el que remarcábamos las implicaciones
de los parámetros sobre las dimensiones de la tabla sobre los tipos de errores.

## Ejecución del benchmark y su entorno
Antes de comenzar a analizar los resultados de la aplicación sintética y extraer información a partir de los datos, pasaremos
a hacer algunas consideraciones sobre la ejecución del mismo que pueden influenciar los datos, de forma que el
posterior análisis sea completo.

### Tipos de ejecuciones: trabajo variable y trabajo constante
Con esta aplicación sintética se han realizado varios benchmarks. En primer lugar, el benchmark era ejecutado de forma que
el valor que se pasaba como número de iteraciones no era el número de iteraciones total que se ejecutaría como suma de
todo el trabajo de los hilos, sino que cada hilo ejecutaría el número de iteraciones indicado. A esta forma de ejecución
la denominaremos "de carga variable", pues la carga de trabajo total varía en función del número de hilos.

Con un sencillo script en bash, se ejecutó el benchmark con 100.000 iteraciones para cada hilo y aumentando progresivamente el resto de dimensiones.
El número de hilos subió desde 4 hasta 256 subiendo de cuatro en cuatro, y las dos dimensiones de la tabla variaron entre 16 y 512
aumentando en potencias de 2. De esta forma obtuvimos información sobre el comportamiento en cualquier combinación
de estas variables dentro de estos rangos. El benchmark se repitió en las mismas condiciones cinco veces y los datos con
los que se generan las gráficas mostradas a continuación en cuanto a ejecuciones con carga variable son la media aritmética de estas cinco ejecuciones.

Posteriormente, se añadió una opción al benchmark de forma que el número de iteraciones que se le hace llegar no representa
el trabajo de cada hilo, sino el trabajo total realizado. Por tanto, el valor que se indique como número de iteraciones se
dividirá entre el número de hilos, manteniendo constante la carga de trabajo independientemente del número de hilos entre los
que se divida. A esta forma de ejecución la denominamos "de carga constante".

### Entorno de experimentación
Como ya comentamos, estos experimentos han sido realizados sobre una máquina virtual
proporcionada por la Escuela. Esta máquina tenía instalado un hipervisor que permitía
compartir el hardware entre varios usuarios de forma transparente. Dado que los benchmarks
realizados no se tratan de una simulación, sino que se ejecutan sobre el procesador real,
que es compartido con otros usuarios, puede que algunos datos se hayan visto afectados por
otras tareas, ya sean de otros usuarios, del hipervisor o del propio sistema operativo.

Para minimizar los efectos que puedan tener estas perturbaciones, las ejecuciones se
intentaron realizar en momentos de bajo uso de la máquina, como noches o fines de semana. Sin
embargo, es imposible garantizar que la ejecución no se haya visto afectada en algunos puntos
por otras tareas como las que indicábamos anteriormente. Esto puede explicar ciertos valores atípicos recogidos.

## Caracterización del comportamiento de la ejecución con memoria transaccional
Para caracterizar el comportamiento de la memoria transaccional partiremos de la principal razón que afecta al
rendimiento de este tipo de programas: los fallos que abortan las transacciones. Cuando una transacción es abortada,
se hace necesario repetir la ejecución y, en nuestro caso, esta repetición se realiza con un cerrojo global tradicional.
Por tanto, variaremos los parámetros que afectan a la forma de la tabla para poder caracterizar el
comportamiento de la memoria transaccional.

### Sensibilidad al tamaño de la tabla
Procediendo con el análisis de los datos recabados, comenzaremos estudiando los efectos que produce la variación del
tamaño de la tabla en el comportamiento de la ejecución con memoria transaccional. Hipotéticamente, los cambios en
esta variable deberían reducir el número de conflictos, puesto que al ser aleatorios los accesos a datos y aumentar el
número de entradas del vector, se reduce la probabilidad de que haya más de un hilo al mismo tiempo usando la misma
línea de caché. Concretamente, los fallos que se verían reducidos al aumentar el tamaño de la tabla serían los fallos
de capacidad generados por fallos en la caché.

Podemos ver este efecto cuando duplicamos el tamaño de la tabla, pasando de 32 a 64 elementos. En este caso vemos en las
gráficas siguientes que el speedup aumenta de forma generalizada. Estas gráficas usan datos de ejecuciones de carga variable
con lo que no podemos comparar las ejecuciones con distinto número de hilos entre sí pero sí podemos comparar 

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/variable/chart-tam-32}
    \end{center}
    \caption{Speedup para una tabla de tamaño 32 (pocos hilos, carga variable)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/variable/chart-tam-64}
    \end{center}
    \caption{Speedup para una tabla de tamaño 64 (pocos hilos, carga variable)}
\end{figure}

También vemos el mismo efecto cuando hacemos la misma comparativa con un número mayor de hilos.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/variable/chart-tam-32-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 32 (muchos hilos, carga variable)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/variable/chart-tam-64-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 64 (muchos hilos, carga variable)}
\end{figure}

Las gráficas representan también cómo evoluciona el rendimiento al cambiar la longitud de la lista enlazada, que provoca
un efecto que contrarresta éste y por eso en mayores longitudes de lista enlazada el rendimiento no aumenta tanto o se
reduce. De estas consideraciones hablaremos en el siguiente apartado.

En estas gráficas vemos que para estos tamaños, el overhead que supone la memoria transaccional no nos da un rendimiento
mejor que el uso de un cerrojo tradicional. Sin embargo, con un número mayor de hilos vemos cómo la ejecución con el
cerrojo se bloquea, serializándose y dándonos un rendimiento peor que con memoria transaccional. Vemos que este efecto
se acentúa cuanto mayor sea el número de hilos y también cuanto mayor sea la estructura.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/variable/chart-tam-128}
    \end{center}
    \caption{Speedup para una tabla de tamaño 128 (pocos hilos, carga variable)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/variable/chart-tam-128-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 128 (muchos hilos, carga variable)}
\end{figure}

Cuando volvemos a duplicar el tamaño desde 64 hasta 128, vemos que sí hay una ligera mejora pero entran en acción otros
efectos que tienden a equiparar el rendimiento de ambas soluciones. Si vamos al caso extremo, que para nosotros es la
tabla de tamaño 512, vemos que según vamos subiendo el tamaño de la tabla llega un momento que no se ven mejoras con el
uso de la memoria transaccional, sino que se llega a un límite con un speedup constante de algo más del 0,9. Podemos 
interpretarlo como que el mayor número de filas en la tabla provoca un mayor uso de líneas de caché, con lo que la caché
es invalidada más rápidamente, y la penalización en el rendimiento de las transacciones abortadas hace que el rendimiento
de la memoria transaccional no supere al de los cerrojos, que debido al gran número de fallos de las transacciones, pasa
a ser más eficiente que la memoria transaccional, aunque no sustancialmente.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/variable/chart-tam-512}
    \end{center}
    \caption{Speedup para una tabla de tamaño 512 (pocos hilos, carga variable)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/variable/chart-tam-512-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 512 (muchos hilos, carga variable)}
\end{figure}

Para tener una visión completa de las características de esta ejecución vamos a ver los datos recogidos con carga
constante, donde sí podemos comparar entre sí los resultados con distinto número de hilos, además de que podemos
comparar también la cantidad total de errores gracias a que el trabajo total es siempre el mismo independientemente
del número de hilos entre los que se divida.

Para los mismos tamaños que comentamos con anterioridad, pero con trabajo constante, podemos ver las siguientes
gráficas.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-32}
    \end{center}
    \caption{Speedup para una tabla de tamaño 32 (pocos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-64}
    \end{center}
    \caption{Speedup para una tabla de tamaño 64 (pocos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-32-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 32 (muchos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-64-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 64 (muchos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-512}
    \end{center}
    \caption{Speedup para una tabla de tamaño 512 (pocos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-512-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 512 (muchos hilos, carga constante de 64m iter.)}
\end{figure}

Aunque las gráficas se pueden consultar en el apéndice correspondiente o en el disco adjunto al trabajo, los speedups
para nuestra ejecución constante con 128 millones de iteraciones y 64 millones de iteraciones son sustancialmente los
mismos. Lo mismo sucede para los errores aunque, claro está, cambia el valor absoluto; pero las gráficas nos muestran
la misma forma. Vemos que con estas cantidades de trabajo, aumentar el número de iteraciones quizás sirva para minimizar
el efecto de agentes externos en los tiempos medidos, pero no vemos un comportamiento distinto.

Como comentábamos antes, nos fijaremos en los tamaños pequeños de lista enlazada puesto que después estudiaremos cómo
afecta la variación en este parámetro. Viendo esos datos vemos que con esta mayor carga de trabajo se mantiene la tendencia
de mejora de rendimiento a mayores tamaños de tabla. Con estos datos podemos ver también cómo ganamos en rendimiento a
la ejecución con cerrojos cuando tenemos un alto número de hilos. También vemos el mismo efecto que comentábamos anteriormente
cuando la tabla aumenta mucho de tamaño: los speedups se hacen constantes y cercanos a 1, independientemente del número de hilos.

Si nos fijamos en algunas gráficas de errores, vemos que la cantidad de errores se reduce a medida que incrementa el tamaño
de la tabla, como especulábamos.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-fails-tam-32-big}
    \end{center}
    \caption{Fallos totales para una tabla de tamaño 32 (muchos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-fails-tam-512-big}
    \end{center}
    \caption{Fallos totales para una tabla de tamaño 512 (muchos hilos, carga constante de 64m iter.)}
\end{figure}

En las dos gráficas anteriores vemos como al pasar de una tabla de tamaño 32 a una de 512, el número de fallos baja. Hay que tener
en cuenta que los fallos al aumentar el tamaño de la lista enlazada los comentaremos después, y que esas dos gráficas tienen escala
distinta.

<!-- TODO Fallos conflicto vs fallos capacidad, debería haber más de conflicto -->

### Sensibilidad al número de entradas de la lista enlazada
Ahora pasamos al siguiente parámetro que afecta al rendimiento de la memoria transaccional: el número de entradas de cada
lista enlazada. Como explicábamos, nuestra tabla está formada por un vector de un tamaño determinado, donde cada nodo apunta
a un elemento de una lista enlazada. Por tanto, la variable que estudiaremos ahora es la longitud de cada una de estas listas
enlazadas que cuelgan de los elementos del vector principal.

Queremos comprobar si, como pensamos, a medida que aumente la longitud de la lista enlazada, los hilos estarán más tiempo
usando cada línea de caché hasta encontrar el dato con el que tienen que trabajar, con lo que es más probable que haya
conflictos en una misma línea de caché. Esta situación provocará un aumento de los fallos de conflicto, que abortarán las
transacciones, provocando penalizaciones en el rendimiento.

Lo primero que observamos a este respecto es que con un tamaño de tabla grande de forma que no nos veamos afectados por los
fallos de conflicto, está claro que a medida que aumenta el número de elementos de la lista enlazada se dispara el número
total de fallos.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/128m/chart-fails-tam-512-big}
    \end{center}
    \caption{Fallos totales para una tabla de tamaño 512 (muchos hilos, carga constante de 128m iter.)}
\end{figure}

Este comportamiento se mantiene con un número menor de hilos también.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/128m/chart-fails-tam-512}
    \end{center}
    \caption{Fallos totales para una tabla de tamaño 512 (pocos hilos, carga constante de 128m iter.)}
\end{figure}

Si reducimos el tamaño de la tabla ya no se observan estos patrones, posiblemente por la influencia de los fallos de
capacidad como ya hemos descrito anteriormente. <!-- TODO Demostrar esto --> Podemos concluir en base a los datos nuestra
hipótesis: longitudes mayores de lista enlazada provocan un número mayor de fallos que lastra el rendimiento.


\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-512}
    \end{center}
    \caption{Speedup para una tabla de tamaño 512 (pocos hilos, carga constante de 128m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-512-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 512 (muchos hilos, carga constante de 128m iter.)}
\end{figure}

Aquí vemos cómo nos lastra sobre todo el aumento de los fallos por conflicto. Vemos cómo surgen las limitaciones del hardware,
pues aunque una estructura distinta podría reducir estos fallos, estamos limitados por la capacidad del hardware. No hemos
encontrado mucha información sobre los planes futuros de Intel u otras compañías con respecto a la memoria transaccional, pero
necesitaríamos soluciones como ampliar el caché de nivel 1 o extender el control con memoria transaccional al caché de nivel 2,
que es sustancialmente más grande y reduciría muy notablemente los fallos por capacidad y por conflicto. Como venimos comentando,
esta tecnología es muy reciente y aquí podemos ver sus limitaciones técnicas, pero eso no debe cegarnos frente a las aplicaciones
que a día de hoy podría tener, y las que tendrá en el futuro cuando la tecnología siga mejorando.

### Comparativa con el rendimiento usando cerrojos tradicionales
El tercer análisis que podemos realizar para caracterizar la memoria transaccional, y para poder entender qué aplicaciones
reales podría tener la memoria transaccional, será comparar el rendimiento de éste método frente a la ejecución con
cerrojos de exclusión mutua.

Hay algunos casos en los que vemos una mejora de rendimiento, con un speedup mayor a uno cuando enfrentamos el tiempo de
ejecución de los cerrojos frente a la memoria transaccional. Por ejemplo, como hemos visto, un mayor tamaño de tabla otorga
una ventaja a la memoria transaccional. Esto puede significar que en tablas reales de gran número de entradas podríamos obtener un
beneficio real aplicando memoria transaccional, aunque habría que probar esta situación en más profundidad para afirmarlo
con seguridad. Probablemente nos encontraríamos con el límite técnico del que hablábamos en la sección anterior.

También vemos que hay una oportunidad de mejora de rendimiento cuando el número de hilos aumenta sustancialmente. Por ejemplo,
a continuación vemos dos gráficas con mismo tamaño de tabla, siendo la primera con un número bajo de hilos mientras que la segunda
muestra los datos para una cantidad de hilos mucho mayor. Podemos ver cómo manteniendo todas las demás variables iguales, el
rendimiento sube debido a que los hilos se bloquean entre sí y no pueden aprovechar el paralelismo real del procesador.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-64}
    \end{center}
    \caption{Speedup para una tabla de tamaño 64 (pocos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-64-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 64 (muchos hilos, carga constante de 64m iter.)}
\end{figure}

Observamos este comportamiento también con las tablas de tamaño 16 y 32, pero ya en tamaños superirores como 128, 256 y 512
vemos cómo empezamos a toparnos con el límite técnico de la implementación en el procesador concreto que usamos para las
pruebas.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-128}
    \end{center}
    \caption{Speedup para una tabla de tamaño 128 (pocos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/64m/chart-tam-128-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 128 (muchos hilos, carga constante de 64m iter.)}
\end{figure}

Por tanto, insistimos en que a día de hoy la memoria transaccional no es un bálsamo mágico que nos va a permitir extraer
rendimiento en cualquier carga de trabajo y algoritmo al aplicarlo a un software con cerrojos preexistente. Debemos entender
bien cómo afectan los distintos tipos de fallos para saber cómo se relacionan éstos con el tipo de estructura con la que
estemos trabajando, de forma que podamos aprovechar la memoria transaccional al máximo. Tampoco podemos perder de vista que
la memoria transaccional seguirá evolucionando en el futuro, aumentando su rendimiento al reducir la probabilidad de fallos.

Sin embargo, sí ganamos algo casi automáticamente al aplicar memoria transaccional, y es una gestión de grano fino con la
complejidad de un cerrojo de exclusión mutua global. En las siguientes gráficas vemos como, a pesar de tener el tamaño más
grande de tabla y de carga de trabajo que hemos probado, los speedups se mantienen rondando el 1.

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/128m/chart-tam-512}
    \end{center}
    \caption{Speedup para una tabla de tamaño 512 (pocos hilos, carga constante de 64m iter.)}
\end{figure}

\begin{figure}[H]
    \begin{center}
        \includegraphics[width=0.7\textwidth]{img/128m/chart-tam-512-big}
    \end{center}
    \caption{Speedup para una tabla de tamaño 512 (muchos hilos, carga constante de 64m iter.)}
\end{figure}
